function [f,df,Hf] = evaluate_function(M, dec, X, area, vertex_normal, corner_angle, type, opt)

% Curvature tensor p-norm
if strcmp(type, 'p-norm')
    assert(isfield(opt,'p'), 'No target power');
    p = opt.p;

    te = regge_curvature_tensor(M, X, vertex_normal);
    [te_norm,Mass_tensor,I,J,S] = regge_matrix_norm(te, M, area, corner_angle);
    
    % f = sum(area.*te_norm.^p)^(1/(2*p))/2;
    f = sum(area.*te_norm.^p)^(1/(2*p))/sqrt(8*pi); % normalization so minimum at 1 for sphere

    if nargout >= 2
        df_u = accumarray(M.T(:), vec(repmat((1/3)*(1-p)*area.*te_norm.^p, [1,3])));
        df_tau = sparse(I, J, -p*area.*te_norm.^(p-1).*S, M.ne, M.ne)*te;
        df = [df_u; df_tau];

        Hf_u = sparse(M.T(:,[1 2 3 1 2 3 1 2 3]), M.T(:,[1 2 3 2 3 1 3 1 2]), repmat((2/9)*(1-p)^2*area.*te_norm.^p, [1,9]), M.nv, M.nv);

        Sh = p*area.*te_norm.^(p-1).*S;
        if p > 1
            te_tri = te(abs(M.T2E));
            S_te = [S(:,1).*te_tri(:,1) + S(:,4).*te_tri(:,2) + S(:,6).*te_tri(:,3), ...
                    S(:,4).*te_tri(:,1) + S(:,2).*te_tri(:,2) + S(:,5).*te_tri(:,3), ...
                    S(:,6).*te_tri(:,1) + S(:,5).*te_tri(:,2) + S(:,3).*te_tri(:,3)];
            S2 = [S_te(:,1).^2, S_te(:,2).^2, S_te(:,3).^2, ...
                  S_te(:,1).*S_te(:,2), S_te(:,2).*S_te(:,3), S_te(:,1).*S_te(:,3)];
            S2 = [S2, S2(:,[6 4 5])];

            Sh = Sh + 2*p*(p-1)*area.*te_norm.^(p-2).*S2;
        end
        Hf_tau = sparse(I, J, Sh, M.ne, M.ne);

        Hf = blkdiag(Hf_u, Hf_tau);
        Hf = (Hf + Hf')/2;
    end
end

% Ricci flow
if strcmp(type, 'ricci')
    assert(isfield(opt,'Kt'), 'No target Gaussian curvature');

    K = angle_defect(M, corner_angle);
    dK = K - opt.Kt;
    if isempty(M.idvx_bound) % Make sure it sums to zero
        dK = dK - mean(dK);
    else
        dK(M.idvx_bound) = dK(M.idvx_bound) - sum(dK)/length(M.idvx_bound);
    end

    f = (dK'*(dec.W\dK))/2;

    if nargout >= 2
        df_u = dK;
        df_tau = zeros(M.ne,1);
        df = [df_u; df_tau];

        Hf_u = dec.W;
        Hf_tau = sparse(M.ne, M.ne);
        Hf = blkdiag(Hf_u, Hf_tau);
        Hf = (Hf + Hf')/2;
    end
end

% p-Willmore flow
if strcmp(type, 'p-willmore')
    assert(isfield(opt,'p'), 'No target power');
    p = opt.p;

    [mean_curv,dmean_curv] = regge_mean_curv(M, X, vertex_normal, corner_angle);
    % f = sum(area.*mean_curv.^(2*p))^(1/(2*p))/2;
    f = sum(area.*mean_curv.^(2*p))^(1/(2*p))/sqrt(4*pi); % normalization so minimum at 1 for sphere

    if nargout >= 2
        df_u = accumarray(M.T(:), vec(repmat((1/3)*(1-p)*area.*mean_curv.^(2*p), [1,3])));
        df_tau =-dmean_curv'*(p*area.*mean_curv.^(2*p-1));
        df = [df_u; df_tau];

        Hf_u = sparse(M.T(:,[1 2 3 1 2 3 1 2 3]), M.T(:,[1 2 3 2 3 1 3 1 2]), repmat((2/9)*(1-p)^2*area.*mean_curv.^(2*p), [1,9]), M.nv, M.nv);
        Hf_tau = dmean_curv'*spdiags(p*(2*p-1)*area.*mean_curv.^(2*p-2), 0, M.nf, M.nf)*dmean_curv;
        Hf = blkdiag(Hf_u, Hf_tau);
        Hf = (Hf + Hf')/2;
    end
end

end

function [H,dH] = regge_mean_curv(M, X, vertex_normal, corner_angle)
nf = M.nf;
ne = M.ne;

cot_ang = cot(corner_angle);

% Evaluate discrete curvature
te = regge_curvature_tensor(M, X, vertex_normal);

% Trace in each triangle
te_tri = te(abs(M.T2E));
H =  te_tri(:,1).*(1 - cot_ang(:,1).*cot_ang(:,2))/2 ...
   + te_tri(:,2).*(1 - cot_ang(:,2).*cot_ang(:,3))/2 ...
   + te_tri(:,3).*(1 - cot_ang(:,3).*cot_ang(:,1))/2;

if nargout > 1
    I = repmat((1:nf)', [1,3]);
    J = abs(M.T2E);
    S = [1 - cot_ang(:,1).*cot_ang(:,2), 1 - cot_ang(:,2).*cot_ang(:,3), 1 - cot_ang(:,3).*cot_ang(:,1)]/2;
    dH = sparse(I, J, S, nf, ne);

    if max(abs(H - dH*te)) > 1e-6
        warning('Invalid mean curvature Jacobian.');
    end
end
end

function K = angle_defect(M, corner_angle)
% Output gaussian curvature inside and normal curvature at the boundary

K = 2*pi - accumarray(M.T(:), corner_angle(:));
K(M.idvx_bound) = K(M.idvx_bound) - pi;
end

function [te_norm,Mass,I,J,S] = regge_matrix_norm(te, M, area, corner_angle)
% Full Regge tansor mass matrix
% i.e. t'*M*t = \sum_{ijk} A_{ijk} (t_{ij} S_{ij}^k + t_{jk} S_{jk}^i + t_{ki} S_{ki}^j)^2
% S(:,[1 2 3]) = [<S_{12}^3, S_{12}^3>, <S_{23}^1, S_{23}^1>, <S_{31}^2, S_{31}^2>]
% S(:,[4 5 6]) = [<S_{23}^1, S_{12}^3>, <S_{31}^2, S_{23}^1>, <S_{12}^3, S_{31}^2>]
% S(:,[7 8 9]) = S(:,[6 4 5])

cot_ang = cot(corner_angle);

% Squared-norm in each triangle
S = [cot_ang(:,[1 2 3]).^2.*cot_ang(:,[2 3 1]).^2 + 0.5*(cot_ang(:,[1 2 3]) - cot_ang(:,[2 3 1])).^2 + 1, ...
    (cot_ang(:,[1 2 3]).*cot_ang(:,[3 1 2]) - 0.5).*(cot_ang(:,[2 3 1]).^2 + 1), ...
    (cot_ang(:,[2 3 1]).*cot_ang(:,[3 1 2]) - 0.5).*(cot_ang(:,[1 2 3]).^2 + 1)];

te_tri = te(abs(M.T2E));
te_norm =  te_tri.^2.*S(:,[1 2 3]) ...
         + te_tri(:,[1 2 3]).*te_tri(:,[2 3 1]).*S(:,[4 5 6]) ...
         + te_tri(:,[1 2 3]).*te_tri(:,[3 1 2]).*S(:,[7 8 9]);
te_norm = sum(te_norm,2);

if nargout > 1
    I = abs(M.T2E(:,[1 2 3 2 3 1 3 1 2]));
    J = abs(M.T2E(:,[1 2 3 1 2 3 1 2 3]));
    S = [cot_ang(:,[1 2 3]).^2.*cot_ang(:,[2 3 1]).^2 + 0.5*(cot_ang(:,[1 2 3]) - cot_ang(:,[2 3 1])).^2 + 1, ...
        (cot_ang(:,[1 2 3]).*cot_ang(:,[3 1 2]) - 0.5).*(cot_ang(:,[2 3 1]).^2 + 1), ...
        (cot_ang(:,[2 3 1]).*cot_ang(:,[3 1 2]) - 0.5).*(cot_ang(:,[1 2 3]).^2 + 1)];
    Mass = sparse(I, J, area.*S, M.ne, M.ne);

    if abs(sum(area.*te_norm) - te'*Mass*te) > 1e-6
        warning('Invalid mass matrix Jacobian.');
    end
end
end