clear all; close all; clc;
cd(fileparts(matlab.desktop.editor.getActiveFilename)); % Go to /Code/ folder
addpath(genpath(pwd));

%% Load mesh
model = 'spot_2'; % Models are stored in ../Data/model
[X,T] = readObj(['../Data/', model, '/', model, '.obj']);

% Flow options
if_save         = false;    % Save meshes in Results folder
if_backtrack    = true;     % Do backtracking step
if_vertex_const = false;    % Add vertex constraints stored in ['../Data/', model, '/idvx_lock.mat']
if_area_const   = false;    % Add total area preservation constraint
fct_type        = 'p-norm'; % Choose energy: 'p-norm' 'ricci' 'p-willmore'
opt.p           = 1;        % Power p for 'p-norm' and 'p-willmore'

% Recenter mesh
[X,area0] = recenter_mesh(X, T);
X0 = X;

% Connectivity information
M = MeshInfo(T);

% Prescribed Gaussian curvature for Ricci flow
if strcmp(fct_type, 'ricci')
    % Compute a cone metric
    [opt.Kt,opt.idvx_cone] = cone_curvature(X, T, true);
end

% Prescribed vertex position
if if_vertex_const
    % Pick data
    try
        load(['../Data/', model, '/idvx_lock.mat']);
    catch
        error('Missing vertex constraint data');
    end
    
    figure(1);
    trisurf(M.T, X(:,1), X(:,2), X(:,3), 1, 'edgecolor', 'k');
    hold on; plot3(X(idvx_lock,1), X(idvx_lock,2), X(idvx_lock,3), 'o', 'MarkerSize', 6, 'MarkerFaceColor', 'r'); 
    text(X(idvx_lock,1), X(idvx_lock,2), X(idvx_lock,3), num2cell(1:length(idvx_lock))); hold off;
    axis equal; colorbar;
    title('Vertex constraints');
else
    idvx_lock = [];
end

% Setup path for saving meshes
if if_save
	if strcmp(fct_type, 'p-willmore') || strcmp(fct_type, 'p-norm')
	    save_path = ['../Data/', model, '/', fct_type, '_', num2str(opt.p), '/'];
	else
	    save_path = ['../Data/', model, '/', fct_type, '/'];
	end
	
    if ~exist(save_path, 'dir')
        mkdir(save_path);
    end
    writeObj([save_path, model, '_', num2str(0,'%04.f'), '.obj'], X, M.T);
end

%% Loop
limit_ang = 25; % step size limitation (in degree)
rho = 1e-4; % Penalty
itmax = 60; % Maximum iteration

opts = optimset('quadprog'); 
opts.ConvexCheck = 'off';

fct = zeros(itmax,1); % Objective function
reset_iteration = false(itmax,1); % Store backtracking steps
time_start = cputime;
timing = zeros(itmax,1);
for it = 1:itmax
    disp(['Iteration: ', num2str(it)]);
    
    % Compute geometric information
    [normal,area,vertex_normal] = compute_normals(X, M.T);
    corner_angle = angles_of_triangles(X, M.T);
    dec = compute_dec_mat(M, corner_angle, area);
    harmonic_forms = primal_harmonic_one_form_basis(M, dec);

    % Compute energy
    [fct(it),df,Hf] = evaluate_function(M, dec, X, area, vertex_normal, corner_angle, fct_type, opt);
    I = build_mass_matrix(M, area);
    norm_df = sqrt(df'*(I\df));
    A = Hf + rho*norm_df*I;
    disp(['Energy: ', num2str(fct(it))]);

    % Build constraint matrix
    Omega = omega_mat(M, X, normal, corner_angle);
    [C,c] = build_flow_constraint(M, dec, X, normal, area, corner_angle, harmonic_forms, Omega, if_area_const, idvx_lock);

    % Find descent direction
    x = quadprog(A, df, [], [], C, c, [],[],[], opts);
    assert(norm(C*x - c) < 1e-5, 'Optimization failed.');
    
    % Limit step size
    om = reshape(Omega*x(1:M.nv+M.ne), [M.ne,3]);
    max_ang = max(sqrt(sum(om.^2,2)))*180/pi;
    t = min(limit_ang/max_ang, 1);
    disp(['Step size : ', num2str(t)]);

    x = t*x;
    u = x(1:M.nv);
    tau = x(M.nv+1:M.nv+M.ne);
    om = reshape(Omega*x(1:M.nv+M.ne), [M.ne,3]);

    % Compute rotation field
    [e1,e2,e3] = compute_rotation_field(M, om);

    % Reconstruct surface
    [X_new,dX] = build_surface(M, X, dec, u, e1, e2, e3);

    % Backtracking
    if if_backtrack
        [normal,area,vertex_normal] = compute_normals(X_new, M.T);
        corner_angle = angles_of_triangles(X_new, M.T);
        dec = compute_dec_mat(M, corner_angle, area);
        fct_new = evaluate_function(M, dec, X_new, area, vertex_normal, corner_angle, fct_type, opt);
        if fct_new < fct(it)
            X = X_new;
            rho = max(0.9*rho, 1e-4);
        else
            rho = 2*rho;
            reset_iteration(it) = true;
            disp('Discard step...')
        end
    else
        X = X_new;
    end

    % Plot stuff
    Q = quasi_conformal_error(M.T, X0, X);
    Qmean = sum(area0.*Q)/sum(area0);
    disp(['Mean conformal error : ', num2str(Qmean)]);
    
    if ~if_save
        [normal,area,vertex_normal] = compute_normals(X, M.T);
        pt = (X(M.T(:,1),:) + X(M.T(:,2),:) + X(M.T(:,3),:))/3;
        figure(1);
        trisurf(M.T, X(:,1), X(:,2), X(:,3), Q, 'edgecolor', 'k');
        hold on; 
        quiver3(pt(:,1), pt(:,2), pt(:,3), normal(:,1), normal(:,2), normal(:,3)); 
        if if_vertex_const
            plot3(X(idvx_lock,1), X(idvx_lock,2), X(idvx_lock,3), 'o', 'MarkerSize', 6, 'MarkerFaceColor', 'r');
        end
        hold off;
        axis equal; colorbar;
        title('Conformal error');
        
        pause; 
    else
        % Save mesh
        writeObj([save_path, model, '_', num2str(it,'%04.f'), '.obj'], X, M.T);
    end
    timing(it) = cputime - time_start;
end

figure;
plot(fct(1:it));
title('Energy');

if if_save
    save([save_path, 'data.mat'], 'fct', 'reset_iteration', 'idvx_lock', 'timing', 'opt');
end
